#include"simulation.hpp"
#include<omp.h>

/**
 * Burridge-Knopoff-Pad model.
 * \f$\begin{aligned}
    m_u \ddot{u}_j &= k_c(u_{j-1} - 2u_j + u_{j+1})-k_p(u_j - x) - m_u \phi(v + \dot{u_j}) \\
    m_x \ddot{x}&= -c_p\dot{x} + k_p \sum_j(u_j -x) - k_{p0}x
    \end{aligned}
   \f$
 */

Simulation::Simulation(string parameters_path):input_parameters_path(parameters_path)
{
    input_parameters = Parameters(parameters_path);
    //arma::arma_rng::set_seed_random();
    arma::arma_rng::set_seed(input_parameters.seed);
    setupModelVectorsMatrices();
    setupConstants();

}

Simulation::Simulation(string parameters_path, bool friction_parallelized):input_parameters_path(parameters_path)
{
    if(!friction_parallelized)
    {
        cout << "Called with friction_parallized parameter, but it is set to false." << endl;
        cout << "Please run without this parameter if the friction parallization version is not wanted." << endl;
    }
    else
    {
        cout << "Running the friction measurement in parallel for different slider speeds." << endl;
        input_parameters = Parameters(parameters_path);
        //arma::arma_rng::set_seed_random();
        arma::arma_rng::set_seed(input_parameters.seed);
        setupConstants();
        calculateFrictionInParallel();
    }


}

//      START: SET UP VECTORS AND MATRICES       //
void Simulation::setupModelVectorsMatrices()
{

    setupBlockMatrix();
    setupBlockMatrixDerivative();

    setupPadVector();
    setupPadVectorDerivative();


    setupSaveVectorsMatrices();
}

void Simulation::setupBlockMatrix()
{
    block_matrix.set_size(input_parameters.N, 2);
    if(input_parameters.debug_negative_intial_values){
        block_matrix = 2.0*block_matrix.randu() - 1.0;
    }

    else if (input_parameters.debug_only_negative_initial)
    {
        block_matrix = block_matrix.randu() * (-1.0);
    }

    else
    {
        block_matrix.randu();
    }
}

void Simulation::setupBlockMatrixDerivative()
{
    block_matrix_derivative.set_size(input_parameters.N, 2);
    block_matrix_derivative.fill(1.0);
}

void Simulation::setupPadVector()
{
    pad_vector = arma::zeros<arma::rowvec>(2);
}

void Simulation::setupPadVectorDerivative()
{
    pad_vector_derivative = arma::zeros<arma::rowvec>(2);
}

void Simulation::setupSaveVectorsMatrices()
{
    // SAVE PURPOSES //
    number_of_time_steps = input_parameters.max_time / input_parameters.dt;
    number_of_save_steps = floor(input_parameters.max_time / (input_parameters.dt * input_parameters.save_interval_dt));
    number_of_interval_save_step = floor(input_parameters.interval / (input_parameters.dt * input_parameters.save_interval_dt));
    if(input_parameters.loggedBlocks.size() < input_parameters.N && input_parameters.loggedBlocks.size() > 0)
    {
        loggingSpecificBlocks = true;
    }
    if(loggingSpecificBlocks)
    {
        block_position_to_file.set_size(input_parameters.loggedBlocks.size(), number_of_save_steps);
        block_velocity_to_file.set_size(input_parameters.loggedBlocks.size(), number_of_save_steps);
    } else
    {
        block_position_to_file.set_size(input_parameters.N, number_of_save_steps);
        block_velocity_to_file.set_size(input_parameters.N, number_of_save_steps);
    }
    printf("Number of time steps: %i\n",number_of_save_steps);
    printf("Logging specific blocks: %d\n", loggingSpecificBlocks);
    block_position_to_file.print();

    pad_position_to_file.set_size(number_of_save_steps);
    pad_velocity_to_file.set_size(number_of_save_steps);
    if(input_parameters.debug_continuous_slider_speed)
    {
        pad_friction_to_file.set_size(number_of_save_steps);
    }else
    {
        pad_friction_to_file.set_size(number_of_interval_save_step);

    }

}

void Simulation::setupSliderSpeedMode()
{
    if(input_parameters.debug_continuous_slider_speed)
    {
        if(input_parameters.interval == 0)
        {
            updateSliderSpeedDtInterval = 1;
            input_parameters.slider_speed = input_parameters.start_speed_continuous;
            input_parameters.increment = (input_parameters.end_speed_continuous - input_parameters.start_speed_continuous) / number_of_time_steps;
        }
        else
        {
            updateSliderSpeedDtInterval = (int)(input_parameters.interval/input_parameters.dt); // used for updating slider speed
        }

    }else
    {
        updateSliderSpeedDtInterval = (int)(input_parameters.interval/input_parameters.dt); // used for updating slider speed
    }
    printf("Slider speed increment: %.10f\n",input_parameters.increment);
}

//      END: SET UP VECTORS AND MATRICES       //

//      START: CONSTANTS       //

void Simulation::setupConstants()
{
    m_u = input_parameters.m_scale_mass*input_parameters.m_mass_x/input_parameters.N;				// Mass of block
	k_p = input_parameters.m_scale_P*input_parameters.m_k_P0/input_parameters.N;						// Stationary spring between pad and blocks
	m_c_crit = 2*sqrt((input_parameters.m_k_P0+input_parameters.N*k_p)*input_parameters.m_mass_x);		// Critical damping coefficient
	c_p = input_parameters.m_zeta*m_c_crit;								// Damper between pad and surface
    k_c = input_parameters.m_scale_C*input_parameters.N*input_parameters.m_k_P0;

    setupSliderSpeedMode();
}

//      END: CONSTANTS       //

//      START: HELP FUNCTIONS CALCULATION       //

/**
* \f$\phi(y) =
	\begin{cases}
		\begin{aligned}
		&F_{0}[-1,1], \quad &y = 0\\
		&F_{0}\frac{1-\sigma}{1+\frac{|y|}{1-\sigma}}sign(y), \quad &y \neq 0
		\end{aligned}
	\end{cases}\f$
*/

double Simulation::phi(double y)
{

	if (y == 0){
        friction_zero_counter += 1;
        printf("[Friction zero counter: %i]", friction_zero_counter);
        return 0; // Maximum static friction force is handled elsewhere
    }

	return (y < 0 ? -1 : 1)*(1 - input_parameters.m_sigma) / (1 + abs(y) / (1 - input_parameters.m_sigma));

}

void Simulation::setSpeedAfterThreshold()
{
    if(input_parameters.slider_speed > input_parameters.threshold_speed && input_parameters.debug_continuous_slider_speed)
    {
        updateSliderSpeedDtInterval = 1e8;
    }
}


void Simulation::updateSliderSpeed(
    int time_step
)
{
    if(time_step % updateSliderSpeedDtInterval == 0)
    {
        printf("[Slider speed: %.6f]",input_parameters.slider_speed);
        setSpeedAfterThreshold();
        if(!input_parameters.debug_continuous_slider_speed)
        {
            if(time_step % input_parameters.save_interval_dt == 0)
            {
                saveToCsvParallel(pad_friction_to_file, "results/", input_parameters.slider_speed);
            }
            index_pad_friction_to_file = 0;
            cout << input_parameters.slider_speed << endl;
        }
        input_parameters.slider_speed += input_parameters.increment;
    }
}

//      END: HELP FUNCTIONS CALCULATION       //

//      START: F FUNCTIONS NUMERICAL SCHEME       //
/**
 *\f$
	\vec{u_{j}}=
	\begin{bmatrix}
	u_{j} \\ \dot{u_{j}}
	\end{bmatrix}
	\quad
	\dot{\vec{u_{j}}} = f(\vec{u_{j}}) =
	\begin{bmatrix}
	\dot{u_{j}} \\ \frac{1}{m_{u}}[k_c(u_{j-1} - 2u_j + u_{j+1})-k_p(u_j - x) - m_u \phi(v + \dot{u_j})]
	\end{bmatrix}

\f$
 */


void Simulation::function_u(
    const arma::mat& current_block_matrix,
    const arma::rowvec& current_pad_vector,
    arma::mat& current_block_matrix_derivative)
{

    arma::colvec spring_forces = arma::zeros<arma::colvec>(input_parameters.N);

    for (int i = 0; i < input_parameters.N; i++) {

        if(!input_parameters.debug_no_neighbor_springs)
        {
		    // Add contribution from neighboring springs, if not deactivated
            spring_forces(i)
                = k_c*(i == 0          ? 0 : (current_block_matrix(i - 1, 0) - current_block_matrix(i, 0)))
                + k_c*(i == input_parameters.N - 1 ? 0 : (current_block_matrix(i + 1, 0) - current_block_matrix(i, 0)));

        }

        if(!input_parameters.debug_no_stationary_springs)
        {
            // Add contribution from stationary springs, if not deactivated
            spring_forces(i) -= k_p*(current_block_matrix(i, 0) - current_pad_vector(0));
        }

        if(!input_parameters.debug_no_friction)
        {
            // Add friction force 
            spring_forces(i) -= m_u*phi(2*input_parameters.m_alpha*input_parameters.slider_speed + current_block_matrix(i, 1));
        }
	}

    current_block_matrix_derivative.col(0) = current_block_matrix.col(1);
    current_block_matrix_derivative.col(1) = spring_forces / m_u;
}

/**
 * \f$\vec{x}=
	\begin{bmatrix}
	x \\ \dot{x}
	\end{bmatrix}
	\quad
	\dot{\vec{x}} = f(\vec{x}) =
	\begin{bmatrix}
	\dot{x} \\ \frac{1}{m_{x}}[-c_p\dot{x} + k_p \sum_j(u_j -x) - k_{p0}x]
	\end{bmatrix}
    \f$
 */

void Simulation::function_x(
    const arma::mat& current_block_matrix,
    const arma::rowvec& current_pad_vector,
    arma::rowvec& current_pad_vector_derivative)
{

    double pad_second_derivative = 0;

    if(!input_parameters.debug_no_damper)
    {
        pad_second_derivative = calculatePadSupportDamper(current_pad_vector);
    }

    if(!input_parameters.debug_no_stationary_springs && !input_parameters.debug_no_pad)
    {
        pad_second_derivative += calculatePadFriction(current_block_matrix, current_pad_vector);
    }


    pad_second_derivative += -input_parameters.m_k_P0*current_pad_vector(0);

    current_pad_vector_derivative(0) = current_pad_vector(1);
    current_pad_vector_derivative(1) = pad_second_derivative / input_parameters.m_mass_x;

    /* cout << pad_second_derivative << endl;
    current_pad_vector.print("current_pad_vector");
    pad_vector_derivative.print("pad_vector_derivative"); */

}

/**
 * \f$\vec{x}= 
	\begin{bmatrix}
	x \\ \dot{x}
	\end{bmatrix}
	\quad
	\dot{\vec{x}} = f(\vec{x}) = 
	\begin{bmatrix}
	\dot{x} \\ \frac{1}{m_{x}}[-c_p\dot{x} - \phi(\dot{x} - v_{0}) - k_{p0}x]
	\end{bmatrix}
    \f$
 */

// ONE DEGREE-OF-FREEDOM //
void Simulation::function_x_odof(
    const arma::rowvec& current_pad_vector,
    arma::rowvec& current_pad_vector_derivative
){
    double pad_second_derivative = 0;

    if(!input_parameters.debug_no_damper)
    {
        pad_second_derivative = calculatePadSupportDamper(current_pad_vector);
    }

    if(!input_parameters.debug_no_stationary_springs && !input_parameters.debug_no_pad)
    {
        pad_second_derivative += -phi(current_pad_vector(1) + input_parameters.slider_speed);
    }


    pad_second_derivative += -input_parameters.m_k_P0*current_pad_vector(0);

    current_pad_vector_derivative(0) = current_pad_vector(1);
    current_pad_vector_derivative(1) = pad_second_derivative / input_parameters.m_mass_x;
}

double Simulation::calculatePadSupportDamper(arma::rowvec current_pad_vector)
{
    return -c_p*current_pad_vector(1);
}

double Simulation::calculatePadFriction(arma::mat current_block_matrix,
                            arma::rowvec current_pad_vector)
{
    return k_p*arma::sum(current_block_matrix.col(0) - current_pad_vector(0));
}


/**
 * \f$\begin{aligned}
	y^{*}  &= y^{n} + \frac{\Delta t}{2} f(y^{n}),\\
	y^{n+1} &= y^{n} + \Delta t f(y^{*})
	\end{aligned}
    \f$
 * \f$\vec{y}=
	\begin{bmatrix}
	y \\ \dot{y}
	\end{bmatrix}
	\quad
	\dot{\vec{y}} = f(\vec{y}) =
	\begin{bmatrix}
	\dot{y} \\ \ddot{y}
	\end{bmatrix}
    \f$
 */

void Simulation::midpointMethod(int time_step)
{
    arma::mat block_matrix_half_step = block_matrix;
    arma::rowvec pad_vector_half_step = pad_vector;

    function_u(block_matrix, pad_vector, block_matrix_derivative);
    function_x(block_matrix, pad_vector, pad_vector_derivative);

    block_matrix_half_step = block_matrix + 0.5 * input_parameters.dt * block_matrix_derivative;
    pad_vector_half_step = pad_vector + 0.5 * input_parameters.dt * pad_vector_derivative;

    function_u(block_matrix_half_step, pad_vector_half_step, block_matrix_derivative);
    function_x(block_matrix_half_step, pad_vector_half_step, pad_vector_derivative);

    block_matrix = block_matrix + input_parameters.dt * block_matrix_derivative;
    pad_vector = pad_vector + input_parameters.dt * pad_vector_derivative;

    if(time_step % input_parameters.save_interval_dt == 0)
    {
        writeValuesToLoggers(time_step);
    }
}

void Simulation::oneDegreeOfFreedomMidpointMethod(
    int time_step
){
    arma::rowvec pad_vector_half_step = pad_vector;

    function_x_odof(pad_vector, pad_vector_derivative);

    pad_vector_half_step = pad_vector + 0.5 * input_parameters.dt * pad_vector_derivative;

    function_x_odof(pad_vector_half_step, pad_vector_derivative);

    pad_vector = pad_vector + input_parameters.dt * pad_vector_derivative;

    if(time_step % input_parameters.save_interval_dt == 0)
    {
        writeValuesToLoggers(time_step);
    }

}

void Simulation::writeValuesToLoggers(const int& time_step){
    if(!input_parameters.debug_only_write_friction)
    {
        if(loggingSpecificBlocks)
        {
            block_position_to_file.col(index_to_file) = block_matrix(input_parameters.loggedBlocks, block_position_column_index);
            block_velocity_to_file.col(index_to_file) = block_matrix(input_parameters.loggedBlocks, block_velocity_column_index);
        }else
        {
            block_position_to_file.col(index_to_file) = block_matrix.col(0);
            block_velocity_to_file.col(index_to_file) = block_matrix.col(1);
        }
        pad_position_to_file(index_to_file) = pad_vector(0);
        pad_velocity_to_file(index_to_file) = pad_vector(1);
    }
    pad_friction_to_file(index_pad_friction_to_file) = calculatePadFriction(block_matrix, pad_vector);
    index_to_file +=1;
    index_pad_friction_to_file += 1;
}

void Simulation::printMidPointMethod(arma::mat block_matrix, arma::rowvec pad_vector, string explanation_text)
{
    block_matrix.print("Block matrix" + explanation_text + ": ");
    pad_vector.print("Pad vector" + explanation_text + ": ");
}

void Simulation::runMidpointMethod()
{
    double percentageTransformer = 100.0/(double)number_of_time_steps;
    cout << "\n\n" << endl;
    for(int i = 0; i < number_of_time_steps; i++)
    {
        printf("\r[Percentage complete: %.0f%%]",(double)i*percentageTransformer);
        if(!input_parameters.debug_one_degree_freedom_mode)
        {
            midpointMethod(i);
        }else
        {
            oneDegreeOfFreedomMidpointMethod(i);
        }
        updateSliderSpeed(i+1);
    }
    cout << "\n\n" << endl;

    saveToCsv(input_parameters.file_name, "results/");
}

void Simulation::saveToCsv(string filename, string result_path = "results/")
{
    if(!input_parameters.debug_only_write_friction)
    {
        block_position_to_file.save(result_path + filename + "_midpoint_block_position.csv", arma::csv_ascii);
        block_velocity_to_file.save(result_path + filename + "_midpoint_block_velocity.csv", arma::csv_ascii);
        pad_position_to_file.save(result_path + filename + "_midpoint_pad_position.csv", arma::csv_ascii);
        pad_velocity_to_file.save(result_path + filename + "_midpoint_pad_velocity.csv", arma::csv_ascii);
    }
    if(input_parameters.debug_continuous_slider_speed)
    {
        pad_friction_to_file.save(result_path + filename + "_midpoint_pad_friction.csv", arma::csv_ascii);
    }
}

// START: Parallel friction calculation //

void Simulation::midpointMethodParallel(
    int timestep,
    arma::mat& block_matrix_parallel,
    arma::rowvec& pad_vector_parallel,
    arma::mat block_matrix_derivative_parallel,
    arma::rowvec pad_vector_derivative_parallel
)
{
    arma::mat block_matrix_half_step = block_matrix_parallel;
    arma::rowvec pad_vector_half_step = pad_vector_parallel;

    // HERE NOW


    function_u(block_matrix_parallel, pad_vector_parallel, block_matrix_derivative_parallel);
    function_x(block_matrix_parallel, pad_vector_parallel, pad_vector_derivative_parallel);

    block_matrix_half_step = block_matrix_parallel + 0.5 * input_parameters.dt * block_matrix_derivative_parallel;
    pad_vector_half_step = pad_vector_parallel + 0.5 * input_parameters.dt * pad_vector_derivative_parallel;

    function_u(block_matrix_half_step, pad_vector_half_step, block_matrix_derivative_parallel);
    function_x(block_matrix_half_step, pad_vector_half_step, pad_vector_derivative_parallel);

    block_matrix_parallel = block_matrix_parallel + input_parameters.dt * block_matrix_derivative_parallel;
    pad_vector_parallel = pad_vector_parallel + input_parameters.dt * pad_vector_derivative_parallel;
}

void Simulation::setupAndRunFrictionInParallel(
    int time_step,
    double slider_speed)
{
    arma::mat block_matrix_parallel;
    arma::mat block_matrix_derivative_parallel;

    arma::rowvec pad_vector_parallel;
    arma::rowvec pad_vector_derivative_parallel;
    block_matrix_parallel.set_size(input_parameters.N, 2);
    if(input_parameters.debug_negative_intial_values){
        block_matrix_parallel = 2.0*block_matrix_parallel.randu() - 1.0;
    }

    else if (input_parameters.debug_only_negative_initial)
    {
        block_matrix_parallel = block_matrix_parallel.randu() * (-1.0);
    }

    else
    {
        block_matrix_parallel.randu();
    }
    block_matrix_derivative_parallel.set_size(input_parameters.N, 2);
    block_matrix_derivative_parallel.fill(1.0);

    pad_vector_parallel = arma::zeros<arma::rowvec>(2);
    pad_vector_derivative_parallel = arma::zeros<arma::rowvec>(2);

    arma::colvec pad_friction_parallel_to_file;
    pad_friction_parallel_to_file.set_size(time_step);

    for(int i = 0; i < time_step; i++)
    {
        midpointMethodParallel(time_step, block_matrix_parallel, pad_vector_parallel, block_matrix_derivative_parallel, pad_vector_derivative_parallel);
        pad_friction_parallel_to_file(i) = calculatePadFriction(block_matrix_parallel, pad_vector_parallel);
    }

    saveToCsvParallel(pad_friction_parallel_to_file, "results/", slider_speed);
}

void Simulation::saveToCsvParallel(
            const arma::colvec& armadillo_vector,
            const string& result_path,
            const double& slider_speed
        )
{
    armadillo_vector.save(result_path + input_parameters.file_name + "_midpoint_pad_friction" + to_string(slider_speed) + ".csv", arma::csv_ascii);
}

void Simulation::calculateFrictionInParallel()
{
    int number_of_slider_speed_intervals = floor(input_parameters.max_time / input_parameters.interval);
    cout << number_of_slider_speed_intervals;
    double slider_speed_parallel = input_parameters.slider_speed;

    double time_steps_per_slider_speed = input_parameters.interval / input_parameters.dt;

    omp_set_num_threads(4);
    #pragma omp parallel for
    for(int i = 1; i < number_of_slider_speed_intervals + 1; i++)
    {
        slider_speed_parallel = input_parameters.slider_speed * i;
        setupAndRunFrictionInParallel(time_steps_per_slider_speed, slider_speed_parallel);
    }

}

// END: Parallel friction calculation //

//      END: F FUNCTIONS NUMERICAL SCHEME       //

void Simulation::main()
{
    setupModelVectorsMatrices();
}

//      DEBUGGING FUNCTIONS      //

void DebugSimulation::printMatrix(arma::mat matrix, bool array_format = false)
{
    cout << endl;
    if(!array_format)
    {
        matrix.print("Matrix: ");
    }else
    {
        int count = 0;
        arma::mat::iterator it_end = matrix.end();
        for(arma::mat::iterator it = matrix.begin(); it != it_end; ++it)
        {
            cout << it[0] << endl;
            count += 1;
        }
        cout << count << endl;
    }
    cout << endl;
}

void DebugSimulation::printRowVector(arma::rowvec vector){
    cout << endl;
    vector.print("Row vector: ");
    cout << endl;
}

void DebugSimulation::main()
{
    setupModelVectorsMatrices();
    /* printMatrix(block_matrix);
    printRowVector(pad_vector); */
    // function_u(block_matrix, pad_vector);
    // function_x(block_matrix, pad_vector);
    /* printMatrix(block_matrix_derivative);
    printRowVector(pad_vector_derivative); */
    runMidpointMethod();
}

//      START: EXAMPLE       //
//      END: EXAMPLE       //
